/*
 * Copyright (C) 2002 Robert Ernst <robert.ernst@linux-solutions.at>
 *
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See COPYING for GPL licensing information.
 *
 */

#include <qpe/qpeapplication.h>
#include <qtimer.h>
#include <qmessagebox.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qfont.h>
#include <qfontmetrics.h>
#include <qtimer.h>
#include "Board.h"
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

Board::Board(QWidget *parent, const char *name, WFlags f) :
    QWidget(parent, name, f)
{
    m_running = false;
    m_time = 0;
    m_parts = 0;
    m_level = 1;
    m_time_left = 0;
    m_parts_left = 0;
    m_placed = 0;
    m_destroyed = 0;
    m_connected = 0;
    m_not_connected = 0;
    m_filled = 0;
    m_spilled = false;
    m_bonus_enabled = false;
    m_sprites = new SpriteCollection(this);
    m_timer = new QTimer(this);
    connect(m_timer, SIGNAL(timeout()), this, SLOT(timeout()));

    int row;
    int col;
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    m_tile[row][col] = -1;
	}
    }

    int i;
    for (i = 0; i < queuesize; i++) {
	m_queue[i] = -1;
    }
}

Board::~Board()
{
    if (m_sprites) {
	delete m_sprites;
    }
    if (m_timer) {
	delete m_timer;
    }
}

void Board::timeout(void)
{
    repaint(false);
    if (m_time_left > 250) {
	m_time_left = m_time_left - 250;
    } else {
	m_time_left = 0;
    }
    if (getTimeForGame() == 0) {
	gameOver();
    }
}

void Board::showEvent(QShowEvent *event)
{
    QWidget::showEvent(event);
    if (m_running) {
	m_timer->start(250);
    }
    setFocusPolicy(StrongFocus);
}

void Board::hideEvent(QHideEvent *event)
{
    QWidget::hideEvent(event);
    m_timer->stop();
}

void Board::paintEvent(QPaintEvent *event)
{
    QPixmap pixmap(width(), height());
    QPainter painter_pixmap(&pixmap);
    QPainter painter_widget(this);

    int yoffs = height() - rows * tilesize;
    int seconds = getTimeForGame();
    int parts = getPartsForGame();
    int score = getScoreForGame();
    QString s;

    painter_pixmap.fillRect(0, 0, width(), yoffs, backgroundColor());
    painter_pixmap.setPen(QColor(0, 0, 0));
    if (m_time && m_parts) {
	s.sprintf(tr("%02d:%02d:%02d/%d"), seconds / 3600, (seconds / 60) % 60, seconds % 60, parts);
    } else if (m_time) {
	s.sprintf(tr("Time: %02d:%02d:%02d"), seconds / 3600, (seconds / 60) % 60, seconds % 60);
    } else if (m_parts) {
	s.sprintf(tr("Parts: %d"), parts);
    } else {
	s.sprintf(tr("No limits"));
    }
    painter_pixmap.drawText(7, 0, width() / 3, yoffs, AlignLeft | AlignVCenter, s);
    painter_pixmap.drawText(width() / 3 + 7, 0, width() / 3, yoffs,
	AlignLeft | AlignVCenter, tr("Score: %1").arg(score));

    int row;
    int col;
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    m_sprites->drawTile(painter_pixmap, col * tilesize,
		yoffs + row * tilesize, m_tile[row][col]);
	}
    }

    int i;
    for (i = 0; i < queuesize; i++) {
	if (i < getPartsForGame()) {
	    m_sprites->drawSmallTile(painter_pixmap,
		width() - 4 - (smalltilesize + 3) * (queuesize - i),
		(yoffs - smalltilesize) / 2, atQueuePos(i));
	}
	if (i == 0) {
	    painter_pixmap.setPen(QColor(255, 0, 0));
	} else if (i == 1) {
	    painter_pixmap.setPen(QColor(0, 0, 0));
	}
	painter_pixmap.drawRect(width() - 4 - (smalltilesize + 3) * (queuesize - i) - 1,
	    (yoffs - smalltilesize) / 2 - 1, smalltilesize + 2, smalltilesize + 2);
    }
    
    painter_pixmap.flush();
    painter_widget.drawPixmap(0, 0, pixmap);
}

void Board::mousePressEvent(QMouseEvent *event)
{
    if (!m_running) {
	start();
	return;
    }

    int yoffs = height() - rows * tilesize;
    int row = (event->y() - yoffs) / tilesize;
    int col = event->x() / tilesize;

    if (row >= 0 && row < rows & col >= 0 && col < cols) {
	if (m_tile[row][col] < 15) {
	    if (m_tile[row][col] == -1) {
		m_placed++;
	    } else {
		m_destroyed++;
	    }
	    m_tile[row][col] = popQueue();
	    repaint(false);
	    if (m_parts_left) {
		m_parts_left--;
	    }
	    if (getPartsForGame() == 0) {
		gameOver();
	    }
	} else if (m_tile[row][col] == 15) {
	    repaint(false);
	    gameOver();
	}
    }
}

void Board::start(void)
{
    int row;
    int col;

    m_running = false;
    m_timer->stop();
    m_time_left = m_time * 1000;
    m_parts_left = m_parts;
    m_placed = 0;
    m_destroyed = 0;
    m_connected = 0;
    m_not_connected = 0;
    m_filled = 0;
    m_spilled = false;
    m_bonus_enabled = false;
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    m_tile[row][col] = -1;
	}
    }
    row = (rand() % (rows - 4)) + 2; // the start
    m_tile[row][cols - 1] = 15;
    row = (rand() % (rows - 4)) + 2;
    m_tile[row][0] = 1 + (rand() % 3); // the end
    fillQueue();
    m_timer->start(250);
    m_running = true;
    repaint(false);
}

void Board::updateTime(int time)
{
    if (time != m_time) {
	m_time = time;
	start();
    }
}

void Board::updateParts(int parts)
{
    if (parts != m_parts) {
	m_parts = parts;
	start();
    }
}

void Board::updateLevel(int level)
{
    if (level != m_level) {
	m_level = level;
	start();
    }
}

int Board::getTimeForGame(void)
{
    if (m_time) {
	return m_time_left / 1000;
    } else {
	return 24 * 60 * 60 - 1;
    }
}

int Board::getPartsForGame(void)
{
    if (m_parts) {
	return m_parts_left;
    } else {
	return 999;
    }
}

int Board::getScoreForGame(void)
{
    int score = m_placed - m_destroyed + m_connected * 5 - m_not_connected * 5 + m_filled * 10;

    if (m_bonus_enabled) {
	if (m_not_connected == 0) {
	    score = score * 2;
	}
	if (m_spilled == false) {
	    score = score * 2;
	}
    }
    return score > 0 ? score : 0;
}

void Board::fillQueue(void)
{
    int i;

    for (i = 0; i < queuesize; i++) {
	pushQueue();
    }
}

void Board::pushQueue(void)
{
    int i;

    for (i = 1; i < queuesize; i++) {
	m_queue[i - 1] = m_queue[i];
    }

    static int fact[3][16] = { // each row sum must be 100
	{5, 5, 5, 5, 20, 20, 6, 6, 6, 6, 3, 3, 3, 3, 4, 0},
	{3, 3, 3, 3, 17, 17, 7, 7, 7, 7, 5, 5, 5, 5, 6, 0},
	{1, 1, 1, 1, 14, 14, 8, 8, 8, 8, 7, 7, 7, 7, 8, 0},
    };
    int r = rand() % 100;

    for (i = 0; i < 16; i++) {
	if (r < fact[m_level % 3][i]) {	
	    m_queue[queuesize - 1] = i;
	    return;
	}
	r = r - fact[m_level % 3][i];
    }
    m_queue[queuesize - 1] = 14;
}

int Board::popQueue(void)
{
    int rv = m_queue[0];

    pushQueue();
    return rv;
}

int Board::atQueuePos(int pos)
{
    return (pos >= 0 && pos < queuesize) ? m_queue[pos] : -1;
}

void Board::gameOver(void)
{
    bool flag[rows][cols];

    m_timer->stop();
    m_running = false;
    m_bonus_enabled = false;
    checkConnected(flag);
    checkSpilled(flag);
    m_bonus_enabled = true;
    repaint(false);
    QMessageBox::information(this, "Pipeman",
	tr("Game over!\nYou scored %1 points").arg(getScoreForGame()));
}

void Board::checkConnected(bool flag[rows][cols])
{
    int row;
    int col;

    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    flag[row][col] = false;
	}
    }
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    if (m_tile[row][col] == 15) {
		checkPartConnected(flag, row, col);
		break;
	    }
	}
    }
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    if (flag[row][col]) {
		m_connected++;
	    } else if (m_tile[row][col] != -1) {
		m_not_connected++;
		m_tile[row][col] = -1;
		repaint(false);
		usleep(100000);
	    }
	}
    }
}

void Board::checkPartConnected(bool flag[rows][cols], int row, int col)
{
    flag[row][col] = true;
    if (hasLeft(m_tile[row][col]) && col > 0 && !flag[row][col - 1] && hasRight(m_tile[row][col - 1])) {
	checkPartConnected(flag, row, col - 1);
    }
    if (hasRight(m_tile[row][col]) && col < (cols - 1) && !flag[row][col + 1] && hasLeft(m_tile[row][col + 1])) {
	checkPartConnected(flag, row, col + 1);
    }
    if (hasUp(m_tile[row][col]) && row > 0 && !flag[row - 1][col] && hasDown(m_tile[row - 1][col])) {
	checkPartConnected(flag, row - 1, col);
    }
    if (hasDown(m_tile[row][col]) && row < (rows - 1) && !flag[row + 1][col] && hasUp(m_tile[row + 1][col])) {
	checkPartConnected(flag, row + 1, col);
    }
}

void Board::checkSpilled(bool flag[rows][cols])
{
    int row;
    int col;
    
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    if (m_tile[row][col] == 15) {
		m_tile[row][col] = 31;
		m_filled = 1;
		repaint(false);
		break;
	    }
	}
    }
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    flag[row][col] = false;
	}
    }
    while (checkSpilledOneRound(flag)) {
	repaint(false);
	usleep(100000);
    }
}

bool Board::checkSpilledOneRound(bool flag[rows][cols])
{
    bool touched[rows][cols];
    int row;
    int col;

    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    touched[row][col] = false;
	}
    }
    for (row = 0; row < rows; row++) {
	for (col = 0; col < cols; col++) {
	    if (m_tile[row][col] > 15 && !flag[row][col] && !touched[row][col]) {
		flag[row][col] = true;
		bool has_left = false;
		bool has_right = false;
		bool has_up = false;
		bool has_down = false;
		if (hasLeft(m_tile[row][col]) && col > 0) {
		    if (hasRight(m_tile[row][col - 1])) {
			has_left = true;
		    } else {
			m_spilled = true;
			return false;
		    }
		}
		if (hasRight(m_tile[row][col]) && col < (cols - 1)) {
		    if (hasLeft(m_tile[row][col + 1])) {
			has_right = true;
		    } else {
			m_spilled = true;
			return false;
		    }
		}
		if (hasUp(m_tile[row][col]) && row > 0) {
		    if (hasDown(m_tile[row - 1][col])) {
			has_up = true;
		    } else {
			m_spilled = true;
			return false;
		    }
		}
		if (hasDown(m_tile[row][col]) && row < (rows - 1)) {
		    if (hasUp(m_tile[row + 1][col])) {
			has_down = true;
		    } else {
			m_spilled = true;
			return false;
		    }
		}
		if (has_left) {
		    if (m_tile[row][col - 1] < 16) {
			m_tile[row][col - 1] = m_tile[row][col - 1] + 16;
			m_filled++;
			touched[row][col - 1] = true;
		    }
		}
		if (has_right) {
		    if (m_tile[row][col + 1] < 16) {
			m_tile[row][col + 1] = m_tile[row][col + 1] + 16;
			m_filled++;
			touched[row][col + 1] = true;
		    }
		}
		if (has_up) {
		    if (m_tile[row - 1][col] < 16) {
			m_tile[row - 1][col] = m_tile[row - 1][col] + 16;
			m_filled++;
			touched[row - 1][col] = true;
		    }
		}
		if (has_down) {
		    if (m_tile[row + 1][col] < 16) {
			m_tile[row + 1][col] = m_tile[row + 1][col] + 16;
			m_filled++;
			touched[row + 1][col] = true;
		    }
		}
		if (m_filled >= m_connected) {
		    return false;
		}
	    }
	}
    }
    return true;
}

bool Board::hasLeft(int which)
{
    static bool has_left[16] = {
	true, false, false, false,
	true, false, true, true, false, false,
	true, false, true, true,
	true,
	false,
    };

    if (which == -1) {
	return false;
    } else {
	return has_left[which % 16];
    }
}

bool Board::hasRight(int which)
{
    static bool has_right[16] = {
	false, true, false, false,
	true, false, false, false, true, true,
	false, true, true, true,
	true,
	false,
    };

    if (which == -1) {
	return false;
    } else {
	return has_right[which % 16];
    }
}

bool Board::hasUp(int which)
{
    static bool has_up[16] = {
	false, false, true, false,
	false, true, true, false, true, false,
	true, true, true, false,
	true,
	false,
    };

    if (which == -1) {
	return false;
    } else {
	return has_up[which % 16];
    }
}

bool Board::hasDown(int which)
{
    static bool has_down[16] = {
	false, false, false, true,
	false, true, false, true, false, true,
	true, true, false, true,
	true,
	true,
    };

    if (which == -1) {
	return false;
    } else {
	return has_down[which % 16];
    }
}

