/*
 * Copyright (C) 2002 Robert Ernst <robert.ernst@linux-solutions.at>
 *
 * This file may be distributed and/or modified under the terms of the
 * GNU General Public License version 2 as published by the Free Software
 * Foundation and appearing in the file LICENSE.GPL included in the
 * packaging of this file.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See COPYING for GPL licensing information.
 *
 */

#include <qpe/qpemenubar.h>
#include <qpe/config.h>
#include <qapplication.h>
#include <qmainwindow.h>
#include <qpopupmenu.h>
#include <qtimer.h>
#include "Pipeman.h"
#include "Board.h"
#include <stdlib.h>
#include <time.h>

Pipeman::Pipeman(QWidget *parent, const char *name, WFlags f) :
    QMainWindow(parent, name, f)
{
    /* initialize random number generator */
    srand(::time(0));

    /* set windows style and caption */
    setCaption("Pipeman");

    /* create the play field as the central widget */
    m_board = new Board(this);
    setCentralWidget(m_board);
    m_board->show();

    /* create the game/time submenu */
    m_timeMenu = new QPopupMenu(0, "time_menu");
    m_timeMenu->setCheckable(true);
    connect(m_timeMenu, SIGNAL(activated(int)), this, SLOT(updateTime(int)));
    connect(m_timeMenu, SIGNAL(activated(int)), m_board, SLOT(updateTime(int)));
    m_timeMenu->insertItem(tr("Unlimited"), 0);
    m_timeMenu->insertItem(tr("2 minutes"), 120);
    m_timeMenu->insertItem(tr("4 minutes"), 240);
    m_timeMenu->insertItem(tr("6 minutes"), 360);

    /* create the game/parts submenu */
    m_partsMenu = new QPopupMenu(0, "parts_menu");
    m_partsMenu->setCheckable(true);
    connect(m_partsMenu, SIGNAL(activated(int)), this, SLOT(updateParts(int)));
    connect(m_partsMenu, SIGNAL(activated(int)), m_board, SLOT(updateParts(int)));
    m_partsMenu->insertItem(tr("Unlimited"), 0);
    m_partsMenu->insertItem(tr("50"), 50);
    m_partsMenu->insertItem(tr("100"), 100);
    m_partsMenu->insertItem(tr("250"), 250);
    m_partsMenu->insertItem(tr("500"), 500);

    /* create the game/computer level submenu */
    m_levelMenu = new QPopupMenu(0, "level_menu");
    m_levelMenu->setCheckable(true);
    connect(m_levelMenu, SIGNAL(activated(int)), this, SLOT(updateLevel(int)));
    connect(m_levelMenu, SIGNAL(activated(int)), m_board, SLOT(updateLevel(int)));
    m_levelMenu->insertItem(tr("Easy"), 0);
    m_levelMenu->insertItem(tr("Medium"), 1);
    m_levelMenu->insertItem(tr("Hard"), 2);
    
    /* create the game submenu */
    QPopupMenu *gameMenu = new QPopupMenu(0, "game_menu");
    gameMenu->insertItem(tr("New"), m_board, SLOT(start()));
    gameMenu->insertItem(tr("Time"), m_timeMenu);
    gameMenu->insertItem(tr("Parts"), m_partsMenu);
    gameMenu->insertItem(tr("Level"), m_levelMenu);
    gameMenu->insertSeparator();
    gameMenu->insertItem(tr("Close"), this, SLOT(close()));

    /* create the menu */
    QPEMenuBar *menu = new QPEMenuBar(this);
    menu->insertItem(tr("Game"), gameMenu);

    /* read the configuration file */
    readConfig();

    /* check the menu entries of game size and game level submenu */
    m_timeMenu->setItemChecked(m_time, true);
    m_partsMenu->setItemChecked(m_parts, true);
    m_levelMenu->setItemChecked(m_level, true);

    /* start a game */
    m_board->start();
}

Pipeman::~Pipeman()
{
    writeConfig();
}

void Pipeman::readConfig(void)
{
    Config cfg("Pipeman");

    cfg.setGroup("Settings");

    int time = cfg.readNumEntry("Time", 0);
    if (time <= 0) {
	m_time = 0;
    } else if (time <= 120) {
	m_time = 120;
    } else if (time <= 240) {
	m_time = 240;
    } else {
	m_time = 360;
    }
    m_board->updateTime(m_time);

    int parts = cfg.readNumEntry("Parts", 0);
    if (parts <= 0) {
	m_parts = 0;
    } else if (parts <= 50) {
	m_parts = 50;
    } else if (parts <= 100) {
	m_parts = 100;
    } else if (parts <= 250) {
	m_parts = 250;
    } else {
	m_parts = 500;
    }
    m_board->updateParts(m_parts);

    int level = cfg.readNumEntry("Level", 1);
    if (level <= 0) {
	m_level = 0;
    } else if (level >= 2) {
	m_level = 2;
    } else {
	m_level = 1;
    }
    m_board->updateLevel(m_level);
}

void Pipeman::writeConfig(void)
{
    Config cfg("Pipeman");

    cfg.setGroup("Settings");
    cfg.writeEntry("Time", m_time);
    cfg.writeEntry("Parts", m_parts);
    cfg.writeEntry("Level", m_level);
}

void Pipeman::updateTime(int time)
{
    m_timeMenu->setItemChecked(m_time, false);
    m_time = time;
    m_timeMenu->setItemChecked(m_time, true);
    writeConfig();
}

void Pipeman::updateParts(int parts)
{
    m_partsMenu->setItemChecked(m_parts, false);
    m_parts = parts;
    m_partsMenu->setItemChecked(m_parts, true);
    writeConfig();
}

void Pipeman::updateLevel(int level)
{
    m_levelMenu->setItemChecked(m_level, false);
    m_level = level;
    m_levelMenu->setItemChecked(m_level, true);
    writeConfig();
}

